local socket = require"socket"
local lfs = require"lfs"
local debug = require"debug"

module("rld", package.seeall)

local debuggerThread
local events = { BREAK = 1, WATCH = 2 }
local breakpoints = {}
local conditions = {}
local stepInto = false
local stepOver = false
local stepLevel = 0
local stackLevel = 0

local host = "localhost"
local hostPort = 22587

local stackDump
local registeredTables = {}
local rootFile = ""

local liteMode = false

showDebugInfo = false

local function encode(o)
	local s = tostring(o)
	s = string.gsub(s, ";", "%%%%%%CM%%%%%%")   -- %% == "%"
	s = string.gsub(s, "\n", "%%%%%%NL%%%%%%")
	s = string.gsub(s, "\r", "%%%%%%RT%%%%%%")
	s = string.gsub(s, "=", "%%%%%%EQ%%%%%%")
	if (type(o) == "string") then
		s = string.format("%q", s)
	end
	return s
end

local function normalizeFilename(str)
    return string.lower(string.gsub(str, "/", "\\"))
end

local function breakPath(path) 
	local paths = {}
	path = string.gsub(path, "\\", "/")
	for w in string.gfind(path, "[^\/]+") do
		table.insert(paths, w)
	end
	return paths
end

local function mergePaths(path1, path2)
	local paths1 = breakPath(path1)
	local paths2 = breakPath(path2)
	for i, path in ipairs(paths2) do
		if path == ".." then
			table.remove(paths1, table.getn(paths1))
		elseif path ~= "." then
			table.insert(paths1, path)
		end
	end
	return table.concat(paths1, "/")
end

local function formatStackSource(ss)
    if string.find(ss, "@") == 1 then
        ss = string.sub(ss, 2)
        if string.find(ss,":") == nil then
            ss = mergePaths(lfs.currentdir(), ss)
        end
    else
        ss = "[SCRIPT] " .. ss
    end
    return ss
end

local function dumpCallstack()
	local d = ""
	local level = 3
	--registeredTables = {}
	tempMaxIdx = 0
	while true do
		local info = debug.getinfo(level,"Sln")
		if info then
			d = d .. level .. ";"
			if info.what == "C" then
				d = d .. "c\r"
			else
				d = d .. "lua;"..encode(formatStackSource(info.source))..";"..encode(info.name)..";"..info.currentline
				local idx = 1
				while true do
					local n, v = debug.getlocal(level, idx)
					local key = n
					if key == nil then
						key = "temp"..tempMaxIdx;
						tempMaxIdx = tempMaxIdx + 1
					end
					if n then
						d = d ..";l="..encode(n).."="..encode(v)
						if type(v) == "table" then
							registeredTables[string.gsub(tostring(v),": ","_")] = v
						end
					else
						break
					end
					idx = idx + 1
				end
				d = d .. "\r"
			end
			level = level + 1
		else
			break
		end
	end
	stackDump = d
end

local function getTable(key)
	key = string.gsub(key,": ","_")
	if registeredTables[key] then
		local ret = ""
		for k, v in pairs(registeredTables[key]) do
			ret = ret..encode(k).."="..encode(v)..";"
			if type(v) == "table" then
				registeredTables[string.gsub(tostring(v),": ","_")] = v
			end
		end
		ret = string.sub(ret, 1, -2)
		return ret
	else
		return nil
	end
end

local internalGlobals = {}
for _, v in pairs({"string", "xpcall", "package", "tostring",
	"print", "os", "unpack", "require", "getfenv",
	"setmetatable", "next", "assert", "tonumber", "io",
	"rawequal", "collectgarbage", "arg", "getmetatable",
	"module", "socket", "rawset", "hrld", "lfs", "math",
	"debug", "pcall", "table", "newproxy", "type", "coroutine",
	"_G", "select", "gcinfo", "pairs", "rawget", "loadstring",
	"ipairs", "_VERSION", "dofile", "setfenv", "load", "error",
	"loadfile" }) do internalGlobals[v] = true end

local function getGlobals(ignoreInternal)
	local ret = ""
	for k, v in pairs(_G) do
		if not (ignoreInternal == 1 and internalGlobals[tostring(k)]) then
			ret = ret..encode(k).."="..encode(v)..";"
			if type(v) == "table" then
				registeredTables[string.gsub(tostring(v),": ","_")] = v
			end
		end
	end
	ret = string.sub(ret, 1, -2)
	return ret
end

local function dumpBreakpoints(bEncode)
	local s = ""
	for fn, bps in pairs(breakpoints) do
		for ln, v in pairs(bps) do
			if v then
				if bEncode then
					s = s .. encode(fn) .. "\t" .. encode(ln) .. "\n"
				else
					s = s .. fn .. "\t" .. ln .. "\n"
				end
			end
		end
	end
	return s
end


local function saveBreakpoints()
	local bpFile = rootFile .. ".rldconfig"
	local s = dumpBreakpoints()
	local f = assert(io.open(bpFile, "w"))
	if f then
		f:write(s)
		f:close()
	end
end

local function setBreakpoint(file, line, save)
	if save == nil then save = true end
    file = normalizeFilename(file)
	if not breakpoints[file] then
		breakpoints[file] = {}
	end
	breakpoints[file][tonumber(line)] = true	
	if save then
		saveBreakpoints()
	end
end

local function loadBreakpoints()
	local bpFile = rootFile .. ".rldconfig"
	local f = io.open(bpFile, "r")
	if f then
		while true do
			local t = f:read("*line")
			
			if not t then break end
			local _, _, file, line = string.find(t, "^([%w%p]+)\t(%d+)$")
			if file and line then
				setBreakpoint(file, line, false)
			end
		end
		f:close()
	end
end

local function deleteBreakpoint(file, line)
    file = normalizeFilename(file)
	if breakpoints[file] then
		breakpoints[file][line] = nil
	end
end

local function hasBreakpoint(file, line)
    file = normalizeFilename(file)
	return breakpoints[file] and breakpoints[file][line]
end

local function restoreVars(vars)
	if type(vars) ~= 'table' then return end
	local func = debug.getinfo(3, "f").func
	local i = 1
	local written_vars = {}
	while true do
		local name = debug.getlocal(3, i)
		if not name then break end
		debug.setlocal(3, i, vars[name])
		written_vars[name] = true
		i = i + 1
	end
	i = 1
	while true do
		local name = debug.getupvalue(func, i)
		if not name then break end
		if not written_vars[name] then
			debug.setupvalue(func, i, vars[name])
			written_vars[name] = true
		end
		i = i + 1
	end
end

local function captureVars()
	local vars = {}
	local func = debug.getinfo(3, "f").func
	local i = 1
	while true do
		local name, value = debug.getupvalue(func, i)
		if not name then break end
		vars[name] = value
		i = i + 1
	end
	i = 1
	while true do
		local name, value = debug.getlocal(3, i)
		if not name then break end
		vars[name] = value
		i = i + 1
	end
	setmetatable(vars, { __index = getfenv(func), __newindex = getfenv(func) })
	return vars
end



local function debugHook(event, line)
	if event == "call" then
		stackLevel = stackLevel + 1
	elseif event == "return" then
		stackLevel = stackLevel - 1
	else
		local file = debug.getinfo(2, "S").source
        file = formatStackSource(file)
		local vars = captureVars()
        if not liteMode then
            table.foreach(conditions, function (index, value)
                setfenv(value, vars)
                local status, res = pcall(value)
                if status and res then
                    dumpCallstack()
                    coroutine.resume(debuggerThread, events.WATCH, vars, file, line, index)
                end
            end)
        end
		if stepInto or (stepOver and stackLevel <= stepLevel) or hasBreakpoint(file, line) then
			stepInto = false
			stepOver = false
			dumpCallstack()
			coroutine.resume(debuggerThread, events.BREAK, vars, file, line)
            restoreVars(vars)
            if liteMode and (stepInto or stepOver) then
                debug.sethook(debugHook, "lcr")
            else
                debug.sethook()
            end
		end
	end
end

function bp()
	local file = debug.getinfo(2, "S").source
	file = formatStackSource(file)
	local line = debug.getinfo(2, "l").currentline
	local vars = captureVars()
	stepInto = false
	stepOver = false
	dumpCallstack();
	coroutine.resume(debuggerThread, events.BREAK, vars, file, line)
    restoreVars(vars)
    if liteMode and (stepInto or stepOver) then
        debug.sethook(debugHook, "lcr")
    else
        debug.sethook()
    end
end


local function debuggerLoop(server)
	local command
	local evalEnv = {}
	while true do
		local line, status = server:receive()
        if showDebugInfo then
            print("DEBUG> " .. line)
        end
		command = string.sub(line, string.find(line, "^[A-Z]+"))
		if command == "SETB" then
			local _, _, filename, line = string.find(line, "^[A-Z]+%s+([%w%p]+)%s+(%d+)$")
			if filename and line then
				setBreakpoint(filename, tonumber(line))
				server:send("200 OK\n")
			else
				server:send("400 Bad Request\n")
			end
		elseif command == "DELB" then
			local _, _, filename, line = string.find(line, "^[A-Z]+%s+([%w%p]+)%s+(%d+)$")
			if filename and line then
				deleteBreakpoint(filename, tonumber(line))
				server:send("200 OK\n")
			else
				server:send("400 Bad Request\n")
			end
		elseif command == "EXEC" then
			local _, _, chunk = string.find(line, "^[A-Z]+%s+(.+)$")
			if chunk then 
				local func = loadstring(chunk)
				local status, res
				if func then
					setfenv(func, evalEnv)
					status, res = xpcall(func, debug.traceback)
				end
				res = encode(tostring(res)) .. "\n"
				if status then
					server:send("200 OK " .. string.len(res) .. "\n") 
					server:send(res)
				else
					server:send("401 Error in Expression " .. string.len(res) .. "\n")
					server:send(res)
				end
			else
				server:send("400 Bad Request\n")
			end
		elseif command == "SETW" then
			local _, _, exp = string.find(line, "^[A-Z]+%s+(.+)$")
			if exp then 
				local func = loadstring("return(" .. exp .. ")")
				local newidx = table.getn(conditions) + 1
				conditions[newidx] = func
				server:send("200 OK " .. newidx .. "\n") 
			else
				server:send("400 Bad Request\n")
			end
		elseif command == "DELW" then
			local _, _, index = string.find(line, "^[A-Z]+%s+(%d+)$")
			index = tonumber(index)
			if index then
				conditions[index] = nil
				server:send("200 OK\n") 
			else
				server:send("400 Bad Request\n")
			end
		elseif command == "RUN" then
			server:send("200 OK\n")
			local ev, vars, file, line, watchIdx = coroutine.yield()
			evalEnv = vars
			if ev == events.BREAK then
				server:send("202 Paused " .. encode(file) .. " " .. line .. "\n")
			elseif ev == events.WATCH then
				server:send("203 Paused " .. encode(file) .. " " .. line .. " " .. watchIdx .. "\n")
			else
				file = encode(file) .. "\n"
				server:send("401 Error in Execution " .. string.len(file) .. "\n")
				server:send(file)
			end
		elseif command == "STEP" then
			server:send("200 OK\n")
			stepInto = true
			local ev, vars, file, line, watchIdx = coroutine.yield()
			evalEnv = vars
			if ev == events.BREAK then
				server:send("202 Paused " .. encode(file) .. " " .. line .. "\n")
			elseif ev == events.WATCH then
				server:send("203 Paused " .. encode(file) .. " " .. line .. " " .. watchIdx .. "\n")
			else
				file = encode(file) .. "\n"
				server:send("401 Error in Execution " .. string.len(file) .. "\n")
				server:send(file)
			end
		elseif command == "OVER" then
			server:send("200 OK\n")
			stepOver = true
			stepLevel = stackLevel
			local ev, vars, file, line, watchIdx = coroutine.yield()
			evalEnv = vars
			if ev == events.BREAK then
				server:send("202 Paused " .. encode(file) .. " " .. line .. "\n")
			elseif ev == events.WATCH then
				server:send("203 Paused " .. encode(file) .. " " .. line .. " " .. watchIdx .. "\n")
			else
				file = encode(file) .. "\n"
				server:send("401 Error in Execution " .. string.len(file) .. "\n")
				server:send(file)
			end
        elseif command == "STAK" then
            if stackDump == nil then
                server:send("400 Bad Request\n")
            elseif stackDump == "" then
                server:send("200 OK 4\n")
                server:send("nil\n")
            else
                server:send("200 OK " .. string.len(stackDump) + 1 .. "\n")
                server:send(stackDump .. "\n")
            end
		elseif command == "GETT" then
			local _, _, tableName = string.find(line, "^[A-Z]+%s+(.+)$")
			local tbl = getTable(tableName)
			if tbl == nil then
				server:send("400 Bad Request\n")
			else
                if tbl == "" then
                    tbl = "nil\n"
                else
                    tbl = tableName.."\r"..getTable(tableName).."\n"
                end
				server:send("200 OK "..string.len(tbl).."\n")
				server:send(tbl)
			end
		elseif command == "GETG" then
			local _, _, ignoreInternal = string.find(line, "^[A-Z]+%s+(%d)$")
			if ignoreInternal ~= "0" and ignoreInternal ~= "1" then
				server:send("400 Bad Request\n")
			else
				local ret = getGlobals(tonumber(ignoreInternal))
				if ret == nil then
					server:send("400 Bad Request\n")
				else
                    if ret == "" then
                        ret = "nil\n"
                    else
                        ret = ret .. "\n"
                    end
					server:send("200 OK "..string.len(ret).."\n")
					server:send(ret)
				end
			end
		elseif command == "LSTB" then
			local dump = string.gsub(dumpBreakpoints(true), "\n", "\r") 
            if dump == "" then
                dump = "nil\n"
            else
                dump = dump .. "\n"
            end
            server:send("200 OK " .. string.len(dump) .. "\n")
			server:send(dump)
		else
			server:send("400 Bad Request\n")
		end
	end
end

debuggerThread = coroutine.create(debuggerLoop)

function config(tab)
	if tab.host then
		host = tab.host
	end
	if tab.port then
		hostPort = tab.port
	end
end

function start(lite)
    if lite then
        liteMode = true
    end
	local server = socket.connect(host, hostPort)
	if server then
		_TRACEBACK = function (message) 
			local err = debug.traceback(message)
			server:send("401 Error in Execution " .. string.len(err) .. "\n")
			server:send(err)
			server:close()
			return err
		end
		rootFile = debug.getinfo(2, "S").source
		rootFile = formatStackSource(rootFile)
		loadBreakpoints()
        if not liteMode then
            debug.sethook(debugHook, "lcr")
        else
            debug.sethook()
        end
		return coroutine.resume(debuggerThread, server)
	end
end

